/*
 * This file is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify this file without charge, but are not authorized to
 * license or distribute it to anyone else except as part of a product
 * or program developed by the user.
 * 
 * THIS FILE IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * This file is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY THIS FILE
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even
 * if Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */

#ifndef lint
static char     sccsid[] = "@(#)gio.c	1.21 90/07/09 Copyright 1989 Sun Microsystems";
#endif

/*
 * GUIDE Intermediate Language (GIL) file input / output interface.
 *
 * This file implements the functions required to read and write GIL
 * files.  The GIL syntax is based on Lisp so that they may be read
 * directly into a Lisp or Scheme interpreter.
 *
 * Many of these functions return a string to indicate the begin or
 * end of a special sequence in the GIL syntax, for example, a list.
 * Internally it is common to only use the first character returned.
 * I did it this way so the functions are consistent and to allow 
 * for flexibility in the future.
 */

#include <stdio.h>
#include <ctype.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <strings.h>
#include <varargs.h>
#include <malloc.h>
#include "guide.h"
#include "gio.h"

/*
 * File pointers.
 */
static FILE    *Inp = stdin;
static FILE    *Outp = stdout;

/*
 * Internal state variables.
 */
static char     Indent[12];			/* indent string */
static int      Newline = TRUE;			/* TRUE if on new output line */
static char	Buf[MAXPATHLEN];		/* work buffer */

/*
 * System error message definitions.
 */
extern int      sys_nerr;
extern char    *sys_errlist[];

/*
 * Internal function declarations.
 */
int		gil_version();
void		skip_comments();
char	       *get_token();

/*
 * Return the string representation of a boolean.
 */
char    *
gio_boolean_string(b)
	int             b;
{
	return b ? gio_true_string() : gio_false_string();
}

/*
 * Close the input file.
 */
void
gio_close_input()
{
	if (Inp != stdin) {
		fclose(Inp);
		Outp = stdin;
	}
}

/*
 * Close the output file.
 */
void
gio_close_output()
{
	if (Outp != stdout) {
		fflush(Outp);
		ftruncate(fileno(Outp), ftell(Outp));
		fclose(Outp);
		Outp = stdout;
	}
}

/*
 * Return the comment string.
 */
char	       *
gio_comment_string()
{
	return ";";
}

/*
 * Return the boolean false string.
 */
char	       *
gio_false_string()
{
	return "nil";
}

/*
 * Return the file begin string.
 */
char	       *
gio_file_begin_string()
{
	return "(";
}

/*
 * Return the file end string.
 */
char	       *
gio_file_end_string()
{
	return ")";
}

/*
 * Get a boolean from the input file.  Returns TRUE if successful.
 */
int
gio_get_boolean(b)
	int	       *b;
{
	char	       *p = get_token();

	if (strcmp(p, gio_true_string()) == 0) {
		*b = TRUE;
		return TRUE;
	}
	if (strcmp(p, gio_false_string()) == 0) {
		*b = FALSE;
		return TRUE;
	}
	return FALSE;
}

/*
 * Return whether we are at end-of-file.
 */
int
gio_get_eof()
{
	int		ch;

	while (isspace(ch = fgetc(Inp)))
		;
	if (ch == EOF)
		return TRUE;
	ungetc(ch, Inp);
	return FALSE;
}

/*
 * Get the start of the file.  Returns TRUE if successful.
 */
int
gio_get_file_begin()
{
	int		ch;

	while (isspace(ch = fgetc(Inp)))
		;
	if (ch == *gio_file_begin_string())
		return TRUE;
	ungetc(ch, Inp);
	return FALSE;
}

/*
 * Get the end of the file.  Returns TRUE if successful.
 */
int
gio_get_file_end()
{
	int		ch;

	while (isspace(ch = fgetc(Inp)))
		;
	if (ch == *gio_file_end_string())
		return TRUE;
	ungetc(ch, Inp);
	return FALSE;
}

/*
 * Get a handler from the input file.  Returns TRUE if successful.
 * Sets the string pointer to a buffer allocated with malloc if the string
 * is not empty, otherwise NULL.  Handler could be a quoted string
 * for special language cases (PostScript, Lisp, etc...) or just
 * a simple name for C.
 */
int
gio_get_handler(s)
	char	      **s;
{
#define	INC		32
	int		ch;
	int		lth;			/* string length */
	int		c = 0;			/* string count */

	while (isspace(ch = fgetc(Inp)))
		;
	ungetc(ch, Inp);

	if (ch != *gio_string_begin_string())
		return gio_get_name(s);
	else {
		*s = malloc(lth = INC);
		*s[c++] = fgetc(Inp);
		while (((ch = fgetc(Inp)) != EOF) && (ch != *gio_string_end_string())) {
			if (c + 2 == lth)
				*s = realloc(*s, lth += INC);
			if (ch == '\\')
				ch = fgetc(Inp);
			(*s)[c++] = (char) ch;
		}
		(*s)[c] = *gio_string_end_string();
		(*s)[c+1] = '\0';
		return TRUE;
	}
#undef INC
}

/*
 * Get an integer from the input file.  Returns TRUE if successful.
 */
int
gio_get_integer(i)
	int	       *i;
{
	char	       *p = get_token();
	char	       *q;
	long		l = strtol(p, &q, 10);

	if (p != q) {
		*i = (int) l;
		return TRUE;
	}
	return FALSE;
}

/*
 * Get a keyword from the input file.  Returns TRUE if successful.
 * The keyword is returned in a static buffer.
 */
int
gio_get_keyword(s)
	char	       **s;
{
	*s = get_token();
	return TRUE;
}

/*
 * Get a list as a string from the input file.  Returns TRUE if successful.
 * Sets the string pointer to a buffer allocated with malloc if the string
 * is not empty, otherwise NULL.
 */
int
gio_get_list(s)
	char	      **s;
{
#define	INC		32
	int		ch;
	int		lth;			/* string length */
	int		c;			/* string count */

	while (isspace(ch = fgetc(Inp)))
		;
	if (ch != *gio_list_begin_string()) {
		ungetc(ch, Inp);
		return FALSE;
	}

	if (gio_get_list_end()) {
		*s = NULL;
		return TRUE;
	}

	*s = malloc(lth = INC);
	for (c = 0; (ch = fgetc(Inp)) != *gio_list_end_string(); c++) {
		if (c + 1 == lth)
			*s = realloc(*s, lth += INC);
		if (ch == '\\')
			ch = fgetc(Inp);
		(*s)[c] = (char) ch;
	}
	(*s)[c] = '\0';
	return TRUE;
#undef INC
}

/*
 * Get the start of a list from the input file.  Returns TRUE if successful.
 */
int
gio_get_list_begin()
{
	int		ch;

	while (isspace(ch = fgetc(Inp)))
		;
	if (ch == *gio_list_begin_string())
		return TRUE;
	ungetc(ch, Inp);
	return FALSE;
}

/*
 * Get the end of a list from the input file.  Returns TRUE if successful.
 */
int
gio_get_list_end()
{
	int		ch;

	while (isspace(ch = fgetc(Inp)))
		;
	if (ch == *gio_list_end_string())
		return TRUE;
	ungetc(ch, Inp);
	return FALSE;
}

/*
 * Get a symbol from the input file.  Returns TRUE if successful.
 * Sets the string pointer to a buffer allocated with malloc if the string
 * is not empty, otherwise NULL.
 */
int
gio_get_name(s)
	char	      **s;
{
	char	       *p = get_token();

	if (strcmp(p, gio_false_string()) == 0)
		*s = NULL;
	else {
		*s = malloc(strlen(p) + 1);
		strcpy(*s, p);
	}
	return TRUE;
}

/*
 * Get the start of an object from the input file.  Returns TRUE if successful.
 */
int
gio_get_object_begin()
{
	int		ch;

	while (isspace(ch = fgetc(Inp)))
		;
	if (ch == *gio_object_begin_string())
		return TRUE;
	ungetc(ch, Inp);
	return FALSE;
}

/*
 * Get the end of an object from the input file.  Returns TRUE if successful.
 */
int
gio_get_object_end()
{
	int		ch;

	while (isspace(ch = fgetc(Inp)))
		;
	if (ch == *gio_object_end_string())
		return TRUE;
	ungetc(ch, Inp);
	return FALSE;
}

/*
 * Get a string from the input file.  Returns TRUE if successful.
 * Sets the string pointer to a buffer allocated with malloc if the string
 * is not empty, otherwise NULL.
 */
int
gio_get_string(s)
	char	      **s;
{
#define	INC		32
	int		ch;
	int		lth;			/* string length */
	int		c = 0;			/* string count */

	while (isspace(ch = fgetc(Inp)))
		;
	if (ch != *gio_string_begin_string()) {
		ungetc(ch, Inp);
		return FALSE;
	}

	if (gio_get_string_end()) {
		*s = NULL;
		return TRUE;
	}

	*s = malloc(lth = INC);
	while (((ch = fgetc(Inp)) != EOF) && (ch != *gio_string_end_string())) {
		if (c + 1 == lth)
			*s = realloc(*s, lth += INC);
		if (ch == '\\')
			ch = fgetc(Inp);
		(*s)[c++] = (char) ch;
	}
	(*s)[c] = '\0';
	return TRUE;
#undef INC
}

/*
 * Get the start of a string from the input file.  Returns TRUE if successful.
 */
int
gio_get_string_begin()
{
	int		ch;

	while (isspace(ch = fgetc(Inp)))
		;
	if (ch == *gio_string_begin_string())
		return TRUE;
	ungetc(ch, Inp);
	return FALSE;
}

/*
 * Get the end of a string from the input file.  Returns TRUE if successful.
 */
int
gio_get_string_end()
{
	int		ch;

	if ((ch = fgetc(Inp)) == *gio_string_end_string())
		return TRUE;
	ungetc(ch, Inp);
	return FALSE;
}

/*
 * Write an integer to the output file.
 */
char	       *
gio_integer_string(i)
	int             i;
{
	sprintf(Buf, "%d", i);
	return Buf;
}

/*
 * Return the string representation of a keyword.
 */
char    *
gio_keyword_string(s)
	char	       *s;
{
	return s;
}

/*
 * Return the list begin string.
 */
char	       *
gio_list_begin_string()
{
	return "(";
}

/*
 * Return the list end string.
 */
char	       *
gio_list_end_string()
{
	return ")";
}

/*
 * Return the string representation of a name.
 */
char	       *
gio_name_string(s)
	char	       *s;
{
	return s ? s : gio_false_string();
}

/*
 * Return the object begin string.
 */
char	       *
gio_object_begin_string()
{
	return "(";
}

/*
 * Return the object end string.
 */
char	       *
gio_object_end_string()
{
	return ")";
}

/*
 * Open an input file.  Returns NULL if successful, otherwise an error
 * message.
 */
char           *
gio_open_gil_input(name)
	char           *name;
{
	int	v;
	char	*errmsg;

	/*
	 * If the input file exists and is the correct version, open it
	 * and skip leading comments.
	 */
	if (Inp = fopen(name, "r")) {
		v = gil_version(Inp);
		if ((v != 0) && (v <= G_VERSION)) {
			skip_comments(Inp);
			return NULL;
		} else
			errmsg = "unrecognized file format";
	} else
		errmsg = sys_errlist[errno]; 

	/*
	 * Otherwise, return an error message.
	 */
	sprintf(Buf, "Cannot open %s: %s\n", name, errmsg);
	return Buf;
}

/*
 * Open an output file.  Returns NULL if successful, otherwise an error
 * message.
 */
char           *
gio_open_gil_output(name)
	char           *name;
{
	struct stat	statbuf;
	int		version_written = FALSE;

	/*
	 * If the output file exists, rename it to a backup name and
	 * start a new output file, preserving any leading comments.
	 * Otherwise, start a new output file.
	 */
	if (stat(name, &statbuf) == OK) {
		char		backup[MAXPATHLEN];
		FILE	       *bakp = NULL;
		char	       *comment = gio_comment_string();

		if (Outp = fopen(name, "a")) {
			fclose(Outp);
			sprintf(backup, "%s.BAK", name);
			if ((rename(name, backup) == OK) &&
			    (bakp = fopen(backup, "r")) &&
			    (Outp = fopen(name, "w"))) {
				while (fgets(Buf, sizeof (Buf), bakp) &&
				       strncmp(Buf, comment, strlen(comment)) == 0)
					if (!version_written) {
						fprintf(Outp, "%s%d\n",
							G_VERSION_PREFIX,
							G_VERSION);
						version_written = TRUE;
					} else
						fputs(Buf, Outp);
				fclose(bakp);
				return NULL;
			}
		}
	} else if (Outp = fopen(name, "w")) {
		fprintf(Outp, "%s%d\n", G_VERSION_PREFIX, G_VERSION);
		return NULL;
	}

	/*
	 * Return a message if unsuccessful.
	 */
	sprintf(Buf, "Cannot open %s: %s\n", name, sys_errlist[errno]);
	return Buf;
}

/*
 * Open an output file.  Returns NULL if successful, otherwise an error
 * message.
 */
char           *
gio_open_output(name)
	char           *name;
{
	if (Outp = fopen(name, "w"))
		return NULL;

	sprintf(Buf, "Cannot open %s: %s\n", name, sys_errlist[errno]);
	return Buf;
}

/*
 * fprintf to the current output file.
 */
void
gio_printf(va_alist)
	va_dcl
{
	va_list		args;
	char	       *fmt;

	if (Newline)
		fputs(Indent, Outp);

	va_start(args);
	fmt = va_arg(args, char *);
	Newline = fmt[strlen(fmt) - 1] == '\n';
	vfprintf(Outp, fmt, args);
	va_end(args);
}

/*
 * Write a character to the output file, preceeded by the indent string if we
 * are on a new line.
 */
void
gio_putc(c)
	char            c;
{
	if (Newline)
		fputs(Indent, Outp);
	fputc(c, Outp);
	Newline = c == '\n';
}

/*
 * Write characters to the output file, preceeded by the indent string if we
 * are on a new line.
 */
void
gio_puts(s)
	char           *s;
{
	if (Newline)
		fputs(Indent, Outp);
	fputs(s, Outp);
	Newline = s[strlen(s) - 1] == '\n';
}

/*
 * Write a boolean to the output file.
 */
void
gio_put_boolean(b)
	int             b;
{
	gio_puts(b ? gio_true_string() : gio_false_string());
}

/*
 * Write an integer to the output file.
 */
void
gio_put_integer(i)
	int             i;
{
	sprintf(Buf, "%d", i);
	gio_puts(Buf);
}

/*
 * Write a keyword to the output file.
 */
void
gio_put_keyword(s)
	char		*s;
{
	gio_puts(s);
}

/*
 * Write a handler to the output file.
 */
void
gio_put_handler(s)
	char	       *s;
{
	char	*newstr;
	char	*p;

	if (s && (s[0] == *gio_string_begin_string()) &&
	    (s[strlen(s)-1] == *gio_string_end_string())) {
		newstr = malloc(strlen(s) + 1);
		strcpy(newstr, s);
		newstr[strlen(s)-1] = '\0';

		fputc(*gio_string_begin_string(), Outp);
		for (p = newstr+1; *p; p++) {
			if ((*p == *gio_string_end_string()) || (*p == '\\'))
				fputc('\\', Outp);
			fputc(*p, Outp);
		}
		fputc(*gio_string_end_string(), Outp);
		free(newstr);
	} else
		gio_put_name(s);
}

/*
 * Write a name to the output file.
 */
void
gio_put_name(s)
	char	       *s;
{
	gio_puts(s ? s : gio_false_string());
}

/*
 * Write a string to the output file.
 */
void
gio_put_string(s)
	char           *s;
{
	char	       *p;

	gio_puts(gio_string_begin_string());
	if (s)
		for (p = s; *p; p++) {
			if (*p == *gio_string_end_string())
				fputc('\\', Outp);
			fputc(*p, Outp);
		}
	gio_puts(gio_string_end_string());
}

/*
 * Set the current indent level of the output file.
 */
void
gio_set_indent(level)
	int             level;
{
	register int    i;

	for (i = 0; i != level; i++)
		Indent[i] = '\t';

	Indent[i] = NULL;
}

/*
 * Return the string begin string.
 */
char	       *
gio_string_begin_string()
{
	return "\"";
}

/*
 * Return the string end string.
 */
char	       *
gio_string_end_string()
{
	return "\"";
}

/*
 * Return the string representation of a string.  Truncates the string if
 * it is too long.
 */
char    *
gio_string_string(s)
	char           *s;
{
	char	       *p;
	int		i = 0;

	Buf[i++] = *gio_string_begin_string();
	if (s)
		for (p = s; *p && i < MAXPATHLEN - 4; p++) {
			if (*p == *gio_string_end_string())
				Buf[i++] = '\\';
			Buf[i++] = *p;
		}
	Buf[i++] = *gio_string_end_string();
	Buf[i] = '\0';
	return Buf;
}

/*
 * Return the boolean true string.
 */
char	       *
gio_true_string()
{
	return "t";
}

/*
 * Internal utility functions.
 */

/*
 * Get the next token from the input stream.  It is assumed that we
 * are not at end-of-file.
 */
static char *
get_token()
{
	int		ch;
	int		c = 0;			/* character count */

	while (isspace(ch = fgetc(Inp)))
		;
	do {
		if (c < sizeof (Buf) - 1)
			Buf[c++] = (char) ch;
	} while (!isspace(ch = fgetc(Inp)) && ch != ')');

	ungetc(ch, Inp);
	Buf[c] = '\0';
	return Buf;
}

/*
 * Return whether a file is a GIL file.  Returns non-0 version number
 * if true, otherwise FALSE.  Leaves file positioned at the beginning.
 */
static int
gil_version(fp)
	FILE	       *fp;
{
	char	*s;
	char	*p;
	long	l;
	int	version = 0;

	rewind(fp);
	if (fgets(Buf, sizeof (Buf), fp) && 
	    strncmp(Buf, G_VERSION_PREFIX, strlen(G_VERSION_PREFIX)) == 0) {
		Buf[strlen(Buf)-1] = '\0';
		s = Buf + strlen(G_VERSION_PREFIX);
		l = strtol(s, &p, 10);

		if((s != p) && ((s + strlen(s)) == p))
			version = (int) l;
	}
	rewind(fp);
	return version;
}

/*
 * Skip leading comments in a GIL file.
 */
static void
skip_comments(fp)
	FILE	       *fp;
{
	long		pos;

	for (;;) {
		pos = ftell(fp);
		if (!fgets(Buf, sizeof (Buf), fp) || Buf[0] != *gio_comment_string())
			break;
	}
	fseek(fp, pos, 0);
}
